@@ -22,7 +22,6 @@
 
 #include <errno.h>
 #include <string.h>
-#include <fcntl.h>
 #include <uuid/uuid.h>
 #include <time.h>
 #include <iconv.h>
@@ -112,6 +111,7 @@ struct segment {
 struct bitlk_signature {
 	uint8_t boot_code[3];
 	uint8_t signature[8];
+	uint16_t sector_size;
 } __attribute__ ((packed));
 
 struct bitlk_superblock {
@@ -291,7 +291,7 @@ static int passphrase_to_utf16(struct crypt_device *cd, char *input, size_t inle
 		*out = outbuf;
 	} else {
 		*out = NULL;
-		free(outbuf);
+		crypt_safe_free(outbuf);
 		log_dbg(cd, "Failed to convert passphrase: %s", strerror(errno));
 		r = -errno;
 	}
@@ -415,6 +415,7 @@ void BITLK_bitlk_fvek_free(struct bitlk_fvek *fvek)
 		return;
 
 	crypt_free_volume_key(fvek->vk);
+	free(fvek);
 }
 
 void BITLK_bitlk_vmk_free(struct bitlk_vmk *vmk)
@@ -504,6 +505,13 @@ int BITLK_read_sb(struct crypt_device *cd, struct bitlk_metadata *params)
 		goto out;
 	}
 
+	params->sector_size = le16_to_cpu(sig.sector_size);
+	if (!(params->sector_size == 512 || params->sector_size == 4096)) {
+		log_err(cd, _("Unsupported sector size %" PRIu16 "."), params->sector_size);
+		r = -EINVAL;
+		goto out;
+	}
+
 	/* read GUID and FVE metadata offsets */
 	if (read_lseek_blockwise(devfd, device_block_size(cd, device),
 		device_alignment(device), &sb, sizeof(sb), fve_offset) != sizeof(sb)) {
@@ -722,6 +730,7 @@ int BITLK_dump(struct crypt_device *cd, struct device *device, struct bitlk_meta
 	log_std(cd, "Info for BITLK%s device %s.\n", params->togo ? " To Go" : "", device_path(device));
 	log_std(cd, "Version:      \t%u\n", params->metadata_version);
 	log_std(cd, "GUID:         \t%s\n", params->guid);
+	log_std(cd, "Sector size:  \t%u [bytes]\n", params->sector_size);
 	log_std(cd, "Created:      \t%s", ctime((time_t *)&(params->creation_time)));
 	log_std(cd, "Description:  \t%s\n", params->description);
 	log_std(cd, "Cipher name:  \t%s\n", params->cipher);
@@ -789,8 +798,14 @@ static int get_recovery_key(struct crypt_device *cd,
 	    - 8 groups of 6 divided by '-'
 	    - each part is a number dividable by 11
 	*/
-	if (passwordLen != BITLK_RECOVERY_KEY_LEN)
-		return 0;
+	if (passwordLen != BITLK_RECOVERY_KEY_LEN) {
+                if (passwordLen == BITLK_RECOVERY_KEY_LEN + 1 && password[passwordLen - 1] == '\n') {
+                        /* looks like a recovery key with an extra newline, possibly from a key file */
+                        passwordLen--;
+                        log_dbg(cd, "Possible extra EOL stripped from the recovery key.");
+                } else
+                        return 0;
+        }
 
 	for (i = BITLK_RECOVERY_PART_LEN; i < passwordLen; i += BITLK_RECOVERY_PART_LEN + 1) {
 		if (password[i] != '-')
@@ -1150,6 +1165,9 @@ int BITLK_activate(struct crypt_device *cd,
 		}
 	}
 
+	if (params->sector_size != SECTOR_SIZE)
+		dmd.flags |= CRYPT_ACTIVATE_IV_LARGE_SECTORS;
+
 	r = dm_targets_allocate(&dmd.segment, num_segments);
 	if (r)
 		goto out;
@@ -1170,7 +1188,7 @@ int BITLK_activate(struct crypt_device *cd,
 						segments[i].iv_offset,
 						segments[i].iv_offset,
 						NULL, 0,
-						SECTOR_SIZE);
+						params->sector_size);
 		if (r)
 			goto out;
 
